; History:
;
; v2.0	1.09.2003: Base driver by Jack R. Ellis
; v2.1	4.09.2003: NASM port, detection, LBA, options by Luchezar Georgiev
; v2.2	6.09.2003: Bug fix and cleanup by Jack & Lucho

%define Version '2.2'

;! This program is free software; you can redistribute it and/or modify
;! it under the terms of the GNU General Public License as published by
;! the Free Software Foundation; either version 2 of the License, or
;! (at your option) any later version.
;!	This program is distributed in the hope that it will be useful,
;!	but WITHOUT ANY WARRANTY; without even the implied warranty of
;!	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;!	GNU General Public License for more details.
;! You should have received a copy of the GNU General Public License along
;! with this program; if not, write to the Free Software Foundation, Inc.,
;! 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  (www.gnu.org)

; This is a DOS driver intended to run ONLY a single UltraDMA hard disk
; on chipsets made by VIA, Intel, SiS, ALi or other manufacturers. The
; disk must be the master on the primary IDE channel. A primary "slave"
; is assumed NOT to be present. Other IDE units (CDROM, ZIP) must use
; the secondary IDE channel and are assumed to be run by other drivers.
;
; The disk is assumed to support standard LBA mode (1024 cylinders, 255
; heads, 63 sectors). An "identify drive" command is sent to the disk
; when the driver loads, and the disk is set to its best UltraDMA mode
; not higher than the best controller mode given by command line option
; with mode 2 (ATA-33) as a default. Disks from ATA-33 to ATA-133 may
; be used. The driver does only read or write requests to the primary
; master disk. All other requests (seeks, etc.) are "passed" back to
; the BIOS or some other driver for processing.
;
; Version 2.00 of this driver is changed so that if the user I/O buffer
; is not DWORD-aligned, fails a VDS "lock" or crosses a 64K boundary, a
; request shall be processed through a 4K driver buffer, using UltraDMA
; I/O to and from the buffer. This prevents "passing" any I/O request
; back to the BIOS for execution in slow PIO mode, as in older versions
; of this driver. Not crossing a 64K boundary is required by Intel, SiS
; and ALi chipsets, and DWORD alignment - by the Intel 82801BA[M] ICH2.

;	Usage in CONFIG.SYS: DEVICE[HIGH]=[path]UDMA2.SYS [options]
;	(after XMS/UMB drivers but before disk cache)
; Options:
; UPPERCASE W = enable direct UDMA writes (default: buffered writes)
;	  2-6 = maximum UDMA mode supported by chipset (default: 2)
; Options can be combined. Unknown command line characters are ignored.
; ("W 5", "W5", "5W", "/5/W", "/W5", "-5 -W", "-W-5", "-5W" are same.)
;
; Direct UDMA writes default to "off" as the VT82C686B controller has a
; BUG that CORRUPTS data on non-VDS writes (fixed in the VT8233A/8235).
; Enable at YOUR OWN RISK! If mode is specified, UDMA mode is set to the
; lowest common mode of your disk and controller, else it defaults to 2.
; Modes: 2 (ATA-33), 3 (ATA-44), 4 (ATA-66), 5 (ATA-100), 6 (ATA-133).

; General Program Equations:

BUFSIZ	equ	4096		; Driver I/O buffer size
SECSBUF equ	BUFSIZ/512	; Sectors per I/O buffer
SECSCYL equ	255*63		; LBA sectors per cylinder
SECSHD	equ	63		; LBA sectors per head
RDYTO	equ	7		; 100-msec ready timeout
HERROR	equ	0FF00h		; Hard-error return code
TIMOUT	equ	0BB00h		; Timed-out  return code
BIOSSEG equ	40h		; BIOS segment value
BIOSTMR equ	6Ch		; BIOS "tick" timer offset
BIOSVDS equ	7Bh		; BIOS VDS flag-byte offset
CR	equ	0Dh		; ASCII carriage-return
LF	equ	0Ah		; ASCII line-feed
%define CRLF	CR,LF

; IDE Controller Register Definitions:

CDATA	equ	1F0h		; Data port
CSUBCM	equ	CDATA+1		; Subcommand register
CSECCT	equ	CDATA+2		; I/O sector count
CCMD	equ	CDATA+7		; Command register
CSTAT	equ	CDATA+7		; Status register

; Controller Status and Command Definitions:

BUSY	equ	80h		; Controller is busy
DSKRDY	equ	40h		; Disk is "ready"
FAULT	equ	20h		; Disk has a "fault"
ONCYL	equ	10h		; Disk is "on cylinder"
ERROR	equ	1		; General error flag
LBABITS equ	0E0h		; Fixed MSB LBA bits: LBA = 1, DEV = 0
SETM	equ	3		; Set Mode subcommand
SETF	equ	0EFh		; Set Features command
DRCMD	equ	0C8h		; DMA read command (write is 0CAh)

; Word and Doubleword Layouts:

struc dbt
	.lb	resb	1
	.hb	resb	1
endstruc
struc dwd
	.lw	resw	1
	.hw	resw	1
endstruc

; Device Address (LBA) Packet Layout:

struc	DAP
	.DASize resb	1	; Packet size in bytes (16 for IA-32)
	.DARes1 resb	1	; Reserved, must be 0
	.DANum	resb	1	; Number of blocks to transfer
	.DARes2 resb	1	; Reserved, must be 0
	.DABuf	resd	1	; Address of transfer buffer
	.DABeg	resd	2	; Starting logical block address
endstruc

; DOS "Request Packet" Layout:

struc	RP
	.RPlenU resb	2	; (Unused by us)
	.RPOp	resb	1	; Opcode
	.RPStat resw	1	; Status word
	.RPResN resb	9	; (Unused by us)
	.RPSize resd	1	; Resident driver size
	.RPCmdl resd	1	; Pointer to command line arguments
endstruc

RPERR	equ	8003h		; "Strategy" packet error flags
RPDON	equ	100h		; "Strategy" packet done flag

; DOS Driver Device Header:

	dd	-1		; Link to next device-header block
	dw	0C000h		; Driver "device attributes"
	dw	Strat		; "Strategy" routine offset
	dw	DevInt		; "Device-Interrupt" routine offset
	db	'UDMA$',0	; Driver name

; Resident Driver Variables:

VDS	equ	$		; VDS Parameter Block:
VDSLen	dd	ResEnd		;   Buffer length
VDSOffs dd	0		;   Buffer offset (32-bit!)
VDSSeg	dd	0		;   Buffer segment (LSW) and ID (MSW)
DMAAddr dd	-1		;   Buffer address (32-bit)
DMALen	dd	80000000h	; DMA buffer length & "end" flag
UserBuf dd	0		; "User" buffer pointer and segment
BufBase dd	Buff		; 32-bit buffer base address
PCIAddr dw	0		; PCI Base UDMA Address (set by init)
Sectors db	0		; Remaining I/O sector count
Params	equ	$		; I/O Parameters:
SectCt	db	0		;   Sector count
LBA	dd	0E0000000h	;   Logical Block Address & commands
IOCmd	db	0		;   Command byte
ChkErrs db	0		; "Check I/O errors" flag (1 if so)
VDSLock db	0		; VDS "lock" flag
DirWr	db	2		; "Direct UDMA write" flag (0 = O.K., 2 = NO)
	align	64		; Athlon XP cache line alignment

; Main driver routine, invoked by INT13 calls. At entry, the registers
; contain the parameters of the desired I/O request, as follows:
;
;	AH	Command code
;	AL	I/O sector count (CHS)
;	CH	Lower 8 bits of starting cylinder number (CHS)
;	CL	Starting sector number and 2MSb of cylinder number (CHS)
;	DH	Starting head number (CHS)
;	DL	Unit number
;	ES:BX	I/O buffer address (CHS)
;	DS:SI	Device Address Packet (LBA)

NotUs	popf			; Request NOT for us - reload CPU flags
	db	0EAh		; "Pass" request to BIOS or other drivers
Int13V	dd	-1		; Previous INT13 vector, set by Init
	align	8
Entry	pushf			; Driver entry point - save CPU flags
	cmp	dl,80h		; Is this request for our disk?
	jne	NotUs		; No, exit quick!
	mov	dl,ah		; Copy command code to DL
	and	dl,0BEh		; Mask out LBA (bit 6) and write (bit 0)
	cmp	dl,2		; Is command a read or write?
	mov	dl,80h		; (Reload DL in any event)
	jne	NotUs		; NOT read or write - exit quick!
	popf			; Valid request - discard entry flags
	sti			; Enable CPU interrupts
	cld			; Ensure we do "forward" string commands!
	push	eax		; Save needed CPU registers
	push	cx		; (Unusual order, for exit logic below)
	push	dx
	push	di
	push	ds
	push	es
	push	si
	push	bp
	push	cs		; Set our DS
	pop	ds
	mov	bp,sp		; Point BP to our "stack frame"
	and	byte [bp+22+dbt.lb],0FEh; Reset exiting carry bit (error flag)
	mov	dl,ah		; Save opcode for LBA test below
	and	ah,0BFh		; Mask Int 13 extension bit (6)
	shl	ah,1		; Set DMA command byte
	add	ah,DRCMD-4
	mov	[IOCmd],ah
	shl	dl,1		; New-style LBA request?
	jns	CalcLBA		; No, calculate the LBA from CHS values
	mov	es,[bp+6]	; Get original DS in ES
	mov	al,[es:si+DAP.DANum]; Save I/O sector count
	mov	[Sectors],al
	mov	eax,[es:si+DAP.DABeg]; Save LBA value in I/O parameters
	mov	[LBA],eax
	or	byte [LBA+dwd.hw+dbt.hb],LBABITS; Set DEV and LBA device bits
	les	dx,[es:si+DAP.DABuf]; Get user I/O buffer address
	jmp	short SaveBuf	; Go save address and check alignment
CalcLBA mov	[Sectors],al	; CHS request - save I/O sector count
	mov	si,SECSHD	; Save starting sector in SI
	and	si,cx
	mov	di,dx		; Save starting head in DI
	shr	cl,6		; Get cylinder number in CX
	xchg	cl,ch
	mov	ax,SECSCYL	; Convert cylinder to sectors
	mul	cx
	xchg	ax,di		; Swap low-order and head number
	mov	al,SECSHD	; Convert head to sectors
	mul	ah
	add	di,ax		; Add to cylinder sectors
	adc	dl,dh
	dec	si		; Get starting sector number -1
	add	si,di		; Add to cylinder/head sectors
	adc	dl,dh
	mov	[LBA+dwd.lw],si ; Save LBA value in I/O parameters
	mov	dh,LBABITS	; Set DEV/LBA bits and ensure bits 27-24 are 0
	mov	[LBA+dwd.hw],dx
	mov	dx,bx		; Set buffer offset in DX
SaveBuf mov	[UserBuf+dwd.lw],dx; Save user I/O buffer address
	mov	[UserBuf+dwd.hw],es
	test	dl,3		; Is user buffer DWORD-aligned?
	jnz	UseBuf		; No, use driver-buffered logic below
	mov	al,[DirWr]	; Get "Direct UDMA write" flag
	test	[IOCmd],al	; Write request but direct-UDMA disabled?
	jnz	UseBuf		; Yes, use driver-buffered logic below
	movzx	eax,byte [Sectors]; Clear EAX & set input sector count
	mov	[SectCt],al
	shl	ax,1		; Set VDS and DMA buffer lengths
	mov	[VDSLen+1],ax
	mov	[DMALen+1],ax
	mov	[VDSOffs+dwd.lw],dx; Save buffer address in VDS block	
	mov	ax,es
	mov	[VDSSeg+dwd.lw],ax
	push	byte BIOSSEG	; Point ES to BIOS data
	pop	es
	test	byte [es:BIOSVDS+dbt.lb],020h; Are VDS services active?
	jz	NoVDS		; No, set up 20-bit DMA address
	mov	word [DMAAddr+dwd.hw],-1; Invalidate DMA address
	push	ds		; Point to VDS descriptor block
	pop	es
	mov	di,VDS
	mov	ax,8103h	; Do VDS "lock" on our I/O buffer
	mov	dx,0Ch
	int	4Bh
	jc	UseBuf		; If error, use buffered logic below
	cmp	word [DMAAddr+dwd.hw],-1; Did we get a valid DMA address?
	je	UseBuf		; No, use driver-buffered logic below
	mov	byte [VDSLock],1; Set VDS "lock" flag
DirDMA	mov	eax,[VDSLen]	; Get I/O buffer length -1
	dec	eax
	add	eax,[DMAAddr]	; Calculate ending DMA address
	xor	ax,ax		; Clear low-order ending address
	cmp	eax,[DMAAddr]	; Will this I/O cross a 64K boundary?
	ja	UseBuf		; Yes, use driver-buffered logic below
	call	Ready		; Await controller- and disk-ready
	call	DoIO		; Do DMA input direct to user buffer
	jmp	short GoodIO	; All done - post "success" & exit
NoVDS	shl	eax,4		; No VDS - set 20-bit DMA address
	add	eax,[VDSOffs]
	mov	[DMAAddr],eax
	jmp	DirDMA		; Go do direct DMA to/from user buffer
UseBuf	mov	eax,[BufBase]	; Buffered - set driver buffer address
	mov	[DMAAddr],eax
	call	Ready		; Await controller- and disk-ready
	align	2
Next	movzx	ax,[Sectors]	; Get remaining I/O sector count
	cmp	al,SECSBUF	; Will next I/O be a full buffer?
	jb	NextSC		; No, use remaining sector count
	mov	al,SECSBUF	; Use full-buffer sector count
NextSC	mov	[SectCt],al	; Set next I/O sector count
	shl	ax,1		; Set DMA buffer length
	mov	[DMALen+1],ax
	cmp	byte [IOCmd],DRCMD; Is this a read request?
	jne	Output		; No, use output logic below
	call	DoIO		; Input next block of sectors
	mov	si,Buff		; Set driver & user buffer pointers
	les	di,[UserBuf]
InMov	mov	cx,128		; Move next sector to user's buffer
	cli			; ("rep movsd" needs "cli/sti" on
	rep	movsd		;   some CPUs, but we should not
	sti			;   disable interrupts very long)
	dec	byte [SectCt]	; More sectors left to move?
	jnz	InMov		; Yes, loop back
	mov	[UserBuf+dwd.lw],di; Update user buffer offset
	jmp	short More	; Go check for more I/O sectors
Output	shr	ax,1		; Output - restore sector count
	push	ds		; Set driver buffer address
	pop	es
	mov	di,Buff
	lds	si,[UserBuf]	; Set user buffer address
	align	2
OutMov	mov	cx,128		; Move next sector to driver buffer
	cli			; (see explanation at "InMov" above)
	rep	movsd
	sti
	dec	ax		; More sectors left to move?
	jnz	OutMov		; Yes, loop back
	push	es		; Reload our DS
	pop	ds
	mov	[UserBuf+dwd.lw],si; Update user buffer offset
	call	DoIO		; Output next block of sectors
More	sub	byte [Sectors],SECSBUF; More I/O sectors left to go?
	ja	Next		; Yes, loop back
	align	2
GoodIO	mov	byte [bp+14+dbt.hb],0; Success! Reset exiting AH
Done	pop	bp		; Reload BP and SI (not needed below)
	pop	si
	shr	byte [VDSLock],1; Was I/O buffer "locked" through VDS?
	jnc	Exit		; No, reload remaining registers
	push	ds		; Point to VDS descriptor block
	pop	es
	mov	di,VDS
	mov	ax,8104h	; Do VDS "unlock" on input buffer
	xor	dx,dx
	int	4Bh
Exit	pop	es		; Reload remaining registers and exit
	pop	ds
	pop	di
	pop	dx
	pop	cx
	pop	eax
	iret

; Subroutine to start I/O. This logic exits by "falling through" to
; the "Ready" routine below, nasty but FAST assembly-language code!

DoIO	mov	al,[IOCmd]	; If reading, get DMA "read" command
	add	al,6
	and	al,8
	mov	dx,[PCIAddr]	; Reset DMA command register
	out	dx,al
	xchg	ax,di		; Save DMA "read/write" byte
	inc	dx		; Reset DMA control register
	inc	dx
	mov	al,24h
	out	dx,al
	mov	dx,CSECCT	; Output our 6 I/O parameters
	mov	si,Params
DMAPar	outsb
	inc	dx
	shr	al,1		; Luckily, exactly 6 LSb in AL are used!
	jnz	DMAPar
	xchg	ax,di		; Reload DMA "read/write" byte
	mov	dx,[PCIAddr]	; Set DMA "GO" command bit
	inc	ax
	out	dx,al
	add	dword [LBA],SECSBUF; Update disk logical-block address
	mov	byte [ChkErrs],ERROR; Set "Check I/O Errors" flag

; Subroutine to await controller- and disk-ready

Ready	mov	ax,BIOSSEG	; Point to BIOS timer
	mov	es,ax
	mov	di,BIOSTMR
	mov	cx,[es:di]	; MSW not needed
	add	cx,RDYTO	; Set up ID command time limit
	xchg	ah,[ChkErrs]	; Load/reset "Check I/O Errors" flag
	mov	dx,[PCIAddr]	; Point to DMA control register
	inc	dx
	inc	dx
ChkDMA	cmp	cx,[es:di]	; Has a DMA transfer timed out?
	je	TimedO		; Yes? Post timeout error & exit!
	in	al,dx		; Read controller DMA status
	and	al,1		; Any DMA still in progress?
	jnz	ChkDMA		; Yes, loop back & check again
	dec	dx		; Reset DMA "GO" command bit
	dec	dx
	out	dx,al
	mov	dx,CSTAT	; Point to controller primary status
ChkRdy	cmp	cx,[es:di]	; Too long without becoming ready?
	je	TimedO		; Yes? Post timeout error and exit!
	in	al,dx		; Get controller primary status
	test	al,BUSY		; Is our controller still busy?
	jnz	ChkRdy		; Yes, loop back and keep checking
	not	al		; Get controller status "not" bits
	test	al,DSKRDY+ONCYL ; Is disk not-ready or off-cylinder?
	jnz	ChkRdy		; Yes, loop back and keep checking
	not	al		; Get normal controller status bits
	test	al,FAULT	; Does disk show a "fault"?
	jnz	Failed		; Yes? Post failure error and exit!
	and	al,ah		; If checking them, any I/O errors?
	jnz	Failed		; Yes? Post failure error and exit!
	ret			; All is well - exit
Failed	mov	ax,HERROR	; Hard error! Get error code
	jmp	short BadIO	; Go set up for an error exit
TimedO	mov	ax,TIMOUT	; Timed out! Get error code
BadIO	pop	dx		; Discard subroutine exit address
	mov	[bp+14],ax	; Set error & reset exiting sector count
	inc	byte [bp+22+dbt.lb]; Set exiting carry bit (error!)
	jmp	Done		; Go see about VDS "unlock" & exit
	align	64		; Athlon XP cache line alignment

; Initialization Variables:

Buff	equ	$		; Start of driver I/O buffer
DvrBase dd	0		; 32-bit driver "base address"
VDSInit db	0		; Initialization VDS "lock" flag
MaxMode db	'2'		; Maximum possible UDMA mode (command option)

; "Strategy" routine. This routine saves our DOS "Init" request packet
; for handling below. At entry, ES:BX points to the packet.

Strat	mov	[cs:UserBuf+dwd.lw],bx; Save DOS request-packet address
	mov	[cs:UserBuf+dwd.hw],es
	retf			; Exit - await DOS "Device Interrupt"

; "Device-Interrupt" routine. This routine initializes the driver and
; is dismissed from memory when the "Init" packet is returned to DOS.

DevInt	pushf			; Entry - save used registers
	push	eax
	push	ebx
	push	ecx
	push	edx
	push	si
	push	edi
	push	ds
	push	es
	push	cs		; Set our DS
	pop	ds
	les	bx,[UserBuf]	; Point to DOS request packet
	cmp	byte [es:bx+RP.RPOp],0; Is this an "Init" packet?
	jne	near I_BadP	; No? Post errors and exit quick!
	mov	dx,TTLMsg	; Display our "title" message
	call	putstr
	les	bx,[es:bx+RP.RPCmdl]; Point to the command line tail
GetNext mov	al,[es:bx]	; Get a command line character
	or	al,al		; Terminate if end of line reached
	jz	FindIt
	cmp	al,CR
	je	FindIt
	cmp	al,LF
	je	FindIt
	cmp	al,'W'		; "Enable Direct UDMA Writes" option?
	jne	ChkMode		; No, continue with the next character
	mov	byte [DirWr],0	; Yes, zero "DirWr" flag (allow direct output) 
ChkMode cmp	al,'2'		; Mode digit?
	jb	ChkNext		; No, continue
	cmp	al,'6'
	ja	ChkNext
	mov	[MaxMode],al	; Yes, store the maximum allowable UDMA mode
ChkNext inc	bx
	jmp	GetNext
FindIt	call	find_it		; Find the UDMA controller PCI base port
	jc	near I_Quit	; Not found? Quit!
	mov	[PCIAddr],ax	; Save command register address ( = base)
	mov	dx,CCMD-1	; Issue "Identify Device" command
	mov	al,0A0h
	out	dx,al
	mov	al,0ECh
	call	I_Cmd
	jc	I_DefM		; Error - display "Default" message
	mov	dx,CDATA	; Point to controller PIO data reg.
	mov	cx,53		; Skip ID bytes 0-105
I_Skp1	in	ax,dx
	loop	I_Skp1
	in	ax,dx		; Read ID bytes 106 and 107
	mov	bh,al		; Save "DMA valid" flag byte
	mov	cl,34		; Skip ID bytes 108-175
I_Skp2	in	ax,dx
	loop	I_Skp2
	in	al,dx		; Read ID bytes 176 and 177
	mov	bl,al		; Save "DMA mode" flag byte
	mov	cl,167		; Skip remaining ID data
I_Skp3	in	ax,dx
	loop	I_Skp3
	test	bh,4		; Are UltraDMA flag bits valid?
	jnz	I_Pick		; Yes, select best UltraDMA mode
I_DefM	mov	dx,DFMsg	; Set up to display "Default" message
	jmp	short I_SelM	; Go display msg. & set UDMA mode 2
I_Pick	mov	al,[CurMode]	; Get "current mode" value
	cmp	al,[MaxMode]	; Maximum allowable mode reached?
	jae	I_GotM		; Yes, use mode 2
	cmp	bl,8		; No, will disk do UDMA mode 3?
	jb	I_GotM		; No, use mode 2
	inc	ax		; Set up for UDMA mode 3
	mov	word [DspMode],'44'
	cmp	al,[MaxMode]	; Maximum allowable mode reached?
	jae	I_GotM		; Yes, use mode 3
	cmp	bl,10h		; No, will disk do UDMA mode 4?
	jb	I_GotM		; No, use mode 3
	inc	ax		; Set up for UDMA mode 4
	mov	word [DspMode],'66'
	cmp	al,[MaxMode]	; Maximum allowable mode reached?
	jae	I_GotM		; Yes, use mode 4
	cmp	bl,20h		; No, will disk do UDMA mode 5?
	jb	I_GotM		; No, use mode 4
	inc	ax		; Set up for UDMA mode 5
	mov	dword [DspMode],'100.'
	cmp	al,[MaxMode]	; Maximum allowable mode reached?
	jae	I_GotM		; Yes, use mode 5
	cmp	bl,40h		; No, will disk do UDMA mode 6?
	jb	I_GotM		; No, use mode 5
	inc	ax		; Set up for UDMA mode 6
	mov	word [DspMode+1],'33'
I_GotM	mov	[CurMode],al	; Update "current mode" value
	mov	dx,SMMsg	; Display "Setting mode" message
I_SelM	call	putstr
	mov	dx,CSUBCM	; Set mode-select subcode
	mov	al,SETM
	out	dx,al
	inc	dx
	mov	al,[CurMode]	; Set desired UltraDMA mode value
	add	al,10h
	out	dx,al
	mov	al,SETF		; Issue set-features command to disk
	call	I_Cmd
	jc	near I_SErr	; Error - display message & exit!
	mov	dx,[PCIAddr]	; Reset DMA command register
	xor	eax,eax
	out	dx,al
	inc	dx		; Reset DMA control register
	inc	dx
	mov	al,24h
	out	dx,al
	mov	ax,cs		; Get driver code segment
	mov	[VDSSeg+dwd.lw],ax; Set code segment in VDS block
	shl	eax,4		; Get 20-bit segment address
	push	byte BIOSSEG	; Point ES to BIOS
	pop	es
	test	byte [es:BIOSVDS+dbt.lb],20h; VDS services active?
	jz	I_SetV		; No, set 20-bit virtual addresses
	push	cs		; Point to VDS descriptor block
	pop	es
	mov	di,VDS
	mov	ax,8103h	; "Lock" this driver into memory
	mov	dx,0Ch
	int	4Bh
	jc	I_VErr		; Failed? Display error msg. & exit
	cmp	word [DMAAddr+dwd.hw],-1; Did we REALLY get a 32-bit address?
	je	I_VErr		; No? Display error message & exit
	mov	byte [VDSInit],1; Set initialization VDS "lock" flag
	mov	eax,[DMAAddr]	; Get 32-bit starting driver address
I_SetV	mov	[DvrBase],eax	; Save 32-bit driver "base" address
	add	[BufBase],eax	; Set 32-bit buffer base address
	mov	ax,BUFSIZ-1	; Calculate ending 32-bit buffer addr.
	cwde
	add	eax,[BufBase]
	xor	ax,ax		; Clear low-order ending address
	cmp	eax,[BufBase]	; Does buffer cross a 64K boundary?
	jna	I_SetL		; No, set command-list address
	shr	byte [VDSInit],1; Cannot load! Was driver "locked"?
	jnc	I_BErr		; No, display error message & exit
	push	cs		; Point to VDS descriptor block
	pop	es
	mov	di,VDS
	mov	ax,8104h	; "Unlock" this driver from memory
	xor	dx,dx
	int	4Bh
I_BErr	mov	dx,BEMsg	; Boundary error! Point to message
	jmp	short I_Err	; Go display error message & exit
I_VErr	mov	dx,VEMsg	; VDS "lock" error! Point to message
	jmp	short I_Err	; Go display error message & exit
I_SetL	mov	eax,[DvrBase]	; Set DMA command-list address
	add	eax,byte DMAAddr
	mov	dx,[PCIAddr]
	add	dl,4		; Point to DMA I/O command pointer
	out	dx,eax
	mov	word [VDSLen+dwd.lw],0; Reset lower word of VDS length
	mov	ax,3513h	; Get current INT13 vector
	int	21h
	mov	[Int13V+dwd.lw],bx; Save vector for "passed" requests
	mov	[Int13V+dwd.hw],es
	mov	ax,2513h	; "Hook" this driver into INT13
	mov	dx,Entry
	int	21h
	les	bx,[UserBuf]	; Post driver size & success code
	mov	word [es:bx+RP.RPSize+dwd.lw],ResEnd
	mov	[es:bx+RP.RPSize+dwd.hw],cs
	mov	word [es:bx+RP.RPStat],RPDON
	jmp	short I_Exit	; Go reload registers and exit
I_SErr	mov	dx,SEMsg	; Set-mode error! Point to message
I_Err	call	putstr		; Display error message & suffix
I_Quit	mov	dx,Suffix
	call	putstr
	les	bx,[UserBuf]	; Post 0 size & error status
	mov	word [es:bx+RP.RPSize+dwd.lw],0
	mov	[es:bx+RP.RPSize+dwd.hw],cs
I_BadP	mov	word [es:bx+RP.RPStat],RPERR; "Unknown command"
I_Exit	pop	es		; Reload registers and exit
	pop	ds
	pop	edi
	pop	si
	pop	edx
	pop	ecx
	pop	ebx
	pop	eax
	popf
	retf

; Subroutine to issue initialization commands to our disk

I_Cmd	mov	dx,CCMD		; Issue desired initialization command
	out	dx,al
	mov	ax,BIOSSEG	; Point to BIOS timer
	mov	es,ax
	mov	di,BIOSTMR
	mov	cx,[es:di]	; MSW not needed
	add	cx,RDYTO	; Set up ID command time limit
I_CmdW	cmp	cx,[es:di]	; Has our ID command timed out?
	je	I_CmdE		; Yes, set CPU carry flag & exit
	in	al,dx		; Get controller alternate status
	test	al,BUSY		; Is our controller still busy?
	jnz	I_CmdW		; Yes, loop back and keep checking
	test	al,ERROR	; Did ID command cause any errors?
	jz	I_CmdX		; No, leave carry flag off & exit
I_CmdE	stc			; Error - set CPU carry flag on
I_CmdX	ret			; Exit

; Initialization and Messages:

TTLMsg	db	'UDMA Hard Disk Driver ', Version, CRLF,'$'
DFMsg	db	CRLF,'Default '
SMMsg	db	CRLF,'UDMA mode '
CurMode db	'2, ATA-'
DspMode db	'33. ',CRLF,'$'
BEMsg	db	CRLF,'64K DMA boundary cross buffer$'
SEMsg	db	CRLF,'Set-mode$'
VEMsg	db	CRLF,'VDS lock$'
Suffix	db	' error - driver NOT loaded',CRLF,'$'

hprint	call	_8p		; Stanislav's tightest but cryptic hex output
_8p	xchg	al,ah
_8print call	_4p
_4p	ror	al,4
_4print push	ax
	and	al,0Fh
	cmp	al,0Ah
	sbb	al,69h
	das
	int	29h
	pop	ax
	ret

putstr	push	ax
	mov	ah,9		; display string in DX
	int	21h
	pop	ax
	ret

; Interface byte table - busmaster, legacy mode (bit writable or not):

itable	db	80h,8Ah,0FAh
it_end	equ	$

; Find the UltraDMA controller. Return its base I/O address in AX (CF on error)

find_it mov	dx,cntrlr
	call	putstr
	xor	edi,edi
	mov	ax,0B101h
	int	1Ah		; PCI BIOS v2.0c+ installation check
	cmp	edx,'PCI '	; BIOS OK?
	je	biosok		; yes, continue
	mov	dx,notsup	; no, BIOS too old
	jmp	short finderr
biosok	mov	si,itable
	cld			; ensure SI will be increased
findev	mov	ecx,10100h	; class 1 (storage), subclass 1 (IDE)
	lodsb			; interface byte
	mov	cl,al
	mov	ax,0B103h	; find PCI class code
	push	si
	xor	si,si		; prepare device index (0)
	int	1Ah		; BH: bus, BL: bit 7-3: device, 2-0: function
	pop	si
	jnc	found		; if found, break loop
	cmp	si,it_end	; end of table?
	jne	findev		; no, continue searching
	mov	dx,notfnd	; yes - not found, show error and exit
	jmp	short finderr
found	mov	dx,busstr
	call	putstr
	mov	ax,bx		; get bus/device/function for display
	xchg	ah,al
	call	_8print		; show bus
	mov	dx,devstr
	call	putstr
	rol	ax,5		; move bits 15:8 to 4:0,15:13
	and	al,11111b
	call	_8print		; show device
	mov	dx,funstr
	call	putstr
	rol	ax,3		; move bits 15:13 to 2:0
	and	al,111b
	call	_4print		; show function
	mov	di,4		; get command low byte
	mov	ax,0B108h	; read configuration byte
	push	bx		; bus/device/function may be distroyed
	int	1Ah
	pop	bx
	and	cl,101b		; mask Bus Master and I/O Space bits
	cmp	cl,101b		; asked bits supported?
	je	bitsup		; yes, continue
	mov	dx,busmer	; no, exit
finderr call	putstr
	stc			; error return
	ret
bitsup	mov	di,20h		; yes, get base address register 4
	mov	ax,0B109h	; read configuration word
	int	1Ah
	xchg	cx,ax		; base address now in AX
	and	ax,0FFFCh	; mask non-signinficant 2LSb
	mov	dx,adrstr	; no, continue
	call	putstr
	call	hprint		; show base address
	clc
	ret			; base in AX

notsup	db	CR,'PCI BIOS 2.0c+ required$'
cntrlr	db	'UDMA controller $'
notfnd	db	'not found$'
busstr	db	'found at bus $'
devstr	db	', device $'
funstr	db	', function $'
adrstr	db	', base $'
busmer	db	CRLF,'Bus master$'

	times Buff+BUFSIZ-$ nop ; "Pad" remainder of I/O buffer: don't remove!
				; (forces RAM occupation at load time)
ResEnd	equ	$		; End of resident driver
